#' @title Turn an R script file or knitr / R Markdown report
#'   into a `drake` workflow plan data frame.
#' @export
#' @seealso [drake_plan()], [make()], [plan_to_code()],
#'   [plan_to_notebook()]
#' @description `code_to_plan()`, [plan_to_code()], and
#'   [plan_to_notebook()] together illustrate the relationships
#'   between `drake` plans, R scripts, and R Markdown documents.
#' @details This feature is easy to break, so there are some rules
#'   for your code file:
#'   1. Stick to assigning a single expression to a single target at a time.
#'     For multi-line commands, please enclose the whole command
#'     in curly braces.
#'     Conversely, compound assignment is not supported
#'     (e.g. `target_1 <- target_2 <- target_3 <- get_data()`).
#'   2. Once you assign an expression to a variable,
#'     do not modify the variable any more.
#'     The target/command binding should be permanent.
#'   3. Keep it simple. Please use the assignment operators rather than 
#'     `assign()` and similar functions.
#' @param path a file path to an R script or `knitr` report.
#' @examples
#' plan <- drake_plan(
#'   raw_data = read_excel(file_in("raw_data.xlsx")),
#'   data = raw_data %>%
#'     mutate(Species = fct_inorder(Species)) %>%
#'     select(-X__1),
#'   hist = create_plot(data),
#'   fit = lm(Sepal.Width ~ Petal.Width + Species, data),
#'   strings_in_dots = "literals"
#' )
#' file <- tempfile()
#' # Turn the plan into an R script a the given file path.
#' plan_to_code(plan, file)
#' # Here is what the script looks like.
#' cat(readLines(file), sep = "\n")
#' # Convert back to a drake plan.
#' if (requireNamespace("CodeDepends")){
#'   code_to_plan(file)
#' }
code_to_plan <- function(path){
  assert_pkg("CodeDepends", install = "BiocManager::install")
  # Suppress harmless partial argument match warnings.
  suppressWarnings(
    nodes <- CodeDepends::getInputs(CodeDepends::readScript(path))
  )
  lapply(nodes, node_plan) %>%
    do.call(what = dplyr::bind_rows) %>%
    parse_custom_columns() %>%
    sanitize_plan()
}

node_plan <- function(node){
  tibble::tibble(
    target = deparse(node@code[[2]]),
    command = wide_deparse(node@code[[3]])
  )
}

#' @title Turn a `drake` workflow plan data frame
#'   into a plain R script file.
#' @export
#' @seealso [drake_plan()], [make()], [code_to_plan()],
#'   [plan_to_notebook()]
#' @description `code_to_plan()`, [plan_to_code()], and
#'   [plan_to_notebook()] together illustrate the relationships
#'   between `drake` plans, R scripts, and R Markdown documents.
#'   In the file generated by `plan_to_code()`, every target/command pair
#'   becomes a chunk of code.
#'   Targets are arranged in topological order
#'   so dependencies are available before their downstream targets.
#'   Please note:
#'   1. You are still responsible for loading your project's
#'     packages, imported functions, etc.
#'   2. Triggers disappear.
#' @param plan workflow plan data frame. See [drake_plan()]
#'   for details.
#' @param con a file path or connection to write to.
#' @examples
#' plan <- drake_plan(
#'   raw_data = read_excel(file_in("raw_data.xlsx")),
#'   data = raw_data %>%
#'     mutate(Species = fct_inorder(Species)) %>%
#'     select(-X__1),
#'   hist = create_plot(data),
#'   fit = lm(Sepal.Width ~ Petal.Width + Species, data),
#'   strings_in_dots = "literals"
#' )
#' file <- tempfile()
#' # Turn the plan into an R script a the given file path.
#' plan_to_code(plan, file)
#' # Here is what the script looks like.
#' cat(readLines(file), sep = "\n")
#' # Convert back to a drake plan.
#' if (requireNamespace("CodeDepends")){
#'   code_to_plan(file)
#' }
plan_to_code <- function(plan, con = stdout()){
  writeLines(text = plan_to_text(plan), con = con)
}

#' @title Turn a `drake` workflow plan data frame
#'   into an R notebook,
#' @export
#' @seealso [drake_plan()], [make()], [code_to_plan()],
#'   [plan_to_code()]
#' @description `code_to_plan()`, [plan_to_code()], and
#'   [plan_to_notebook()] together illustrate the relationships
#'   between `drake` plans, R scripts, and R Markdown documents.
#'   In the file generated by `plan_to_code()`, every target/command pair
#'   becomes a chunk of code.
#'   Targets are arranged in topological order
#'   so dependencies are available before their downstream targets.
#'   Please note:
#'   1. You are still responsible for loading your project's
#'     packages, imported functions, etc.
#'   2. Triggers disappear.
#' @inheritParams plan_to_code
#' @examples
#' plan <- drake_plan(
#'   raw_data = read_excel(file_in("raw_data.xlsx")),
#'   data = raw_data %>%
#'     mutate(Species = fct_inorder(Species)) %>%
#'     select(-X__1),
#'   hist = create_plot(data),
#'   fit = lm(Sepal.Width ~ Petal.Width + Species, data),
#'   strings_in_dots = "literals"
#' )
#' file <- tempfile()
#' # Turn the plan into an R notebook a the given file path.
#' plan_to_notebook(plan, file)
#' # Here is what the script looks like.
#' cat(readLines(file), sep = "\n")
#' # Convert back to a drake plan.
#' if (requireNamespace("CodeDepends")){
#'   code_to_plan(file)
#' }
plan_to_notebook <- function(plan, con){
  c(
    "---",
    "title: \"My Notebook\"",
    "output: html_notebook",
    "---",
    "",
    "```{r my_code}",
    plan_to_text(plan),
    "```"
  ) %>%
    writeLines(con = con)
}

plan_to_text <- function(plan){
  . <- NULL
  order <- drake_config(
    plan[, c("target", "command")],
    envir = new.env(parent = emptyenv()),
    cache = storr::storr_environment(),
    verbose = FALSE
  )$graph %>%
    igraph::topo_sort() %>%
    .$name %>%
    intersect(y = plan$target) %>%
    match(table = plan$target)
  plan <- plan[order, ]
  if (!is.character(plan$command)){
    plan$command <- purrr::map_chr(plan$command, rlang::expr_text)
  }
  text <- paste(plan$target, "<-", plan$command)
  if (requireNamespace("styler")){
    text <- styler::style_text(text)
  }
  text
}
